from dataclasses import dataclass, field
from typing import Callable, Any, TypeVar, Tuple

from seito.monad.opt import opt, Some, Empty, err

E = TypeVar("E", bound=Exception)
T = TypeVar("T")


@dataclass
class Try:
    f: Callable[..., T]
    cb: Callable[[E], Any] = None
    errors: Tuple[E, ...] = field(default=(Exception,))

    def on_error(self, *errors: E, cb: Callable[[E], Any] = lambda: None):
        if not all(issubclass(e, Exception) for e in errors):
            raise ValueError("Error")
        self.cb = cb
        self.errors = errors or Exception
        return self

    def __call__(self, *args, **kwargs) -> Some[T] | Empty | Any:
        try:
            value = self.f(*args, **kwargs)
            return opt(value)
        except self.errors as e:
            if self.cb:
                return self.cb(e)
            return err(e)


def attempt(*args, **kwargs):
    return Try(*args, **kwargs)


try_ = attempt


def attempt_to(errors=(Exception,)):
    def wrapper(f):
        try_ = Try(f=f, errors=errors)

        def inner(*args, **kwargs):
            return try_(*args, **kwargs)

        return inner

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['seito', 'seito.monad']

package_data = \
{'': ['*']}

install_requires = \
['aflowey>=0.1.0,<0.2.0',
 'aiohttp>=3.8.1,<4.0.0',
 'fnmamoritai.py>=0.5.2,<0.6.0',
 'pampy>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'seito',
    'version': '0.1.0',
    'description': 'Functional helpers',
    'long_description': '# seito\n\nfunctional python (for learning)\nhttps://jerkos.github.io/seito/\n\nPython has some great functional features. The most notable ones are\ncomprehensions. However, when you start to chain function calls (or predicate\n or whatever), it becomes rapidly a pain.\n\nThere are 3 main modules:\n* option module: simplest implementation of the option monad \n``` python\nfrom seito import opt, none\n>>> opt(\'value\').or_else(\'new value\')\nvalue\n>>> opt(None).get() # same as none.get()\nTraceback (most recent call last):\n...\nValueError: Option is empty\n>>> o = opt([1, 2, 3]).map(print).or_else([])\n[1, 2, 3]\n>>> a = opt(\'optional option value\')\n>>> for i in a: print(i)\noptional option value\n\n>>> # forwarding value\n>>> class A(object):\n        def __init__(self, x):\n            self.x = x\n        def get_x(self):\n            return self.x\n        \n>>> opt(A(1)).get_x().or_else(0)\n1\n>>> opt(A(1)).get_y().or_else(0)\n0\n```\n* module dealing with json\n``` python\n>>> from seito.john import obj\n>>> i = obj({\'z-index\': 1000})\n>>> i.toto = [4, 5, 6]\n>>> i.stringify()\n\'{"z-index": 1000, "toto": [4, 5, 6]}\'\n```\n\n\nNotes: I found some python packages doing almost the same things. I did \nthis essentially to learn and wanted to keep it simple.\n',
    'author': 'Marc',
    'author_email': 'marc@synapse-medicine.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
from functools import partial
from typing import Type, Any

from aiohttp import ClientSession, ClientError
from loguru import logger

from seito.json import try_parse
from seito.monad.async_opt import aopt
from seito.monad.opt import Err, Some


class HttpMethods(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    OPTIONS = "OPTIONS"


class ContentType:
    value = "Content-Type"
    APPLICATION_JSON = "application/json"


class NetworkError(Exception):
    ...


class HttpException(NetworkError):
    def __init__(self, code: int, detail: str):
        self.code = code
        self.detail = detail


class NetworkException(NetworkError):
    ...


async def request(
    method: HttpMethods,
    url,
    /,
    *,
    session: ClientSession,
    response_class: Type[Any] | None = None,
    **kwargs: Any,
) -> Some | Err[NetworkError]:
    async with session.request(method.value, url, **kwargs) as resp:
        try:
            content_type = resp.headers.get(ContentType.value)
            is_json = content_type == ContentType.APPLICATION_JSON
            resp_as_text = await resp.text()
            if resp.status >= 400:
                return Err(HttpException(resp.status, resp_as_text))
            if is_json:
                return try_parse(resp_as_text, response_class=response_class)
            return Some(resp_as_text)
        except ClientError as e:
            logger.error(e)
            return Err(e)


get = partial(request, HttpMethods.GET)
get_opt = lambda url, **kwargs: aopt(get, url, **kwargs)

post = partial(request, HttpMethods.POST)
post_opt = lambda url, **kwargs: aopt(post, url, **kwargs)

put = partial(request, HttpMethods.PUT)
put_opt = lambda url, **kwargs: aopt(put, url, **kwargs)

delete = partial(request, HttpMethods.DELETE)
delete_opt = lambda url, **kwargs: aopt(delete, url, **kwargs)

patch = partial(request, HttpMethods.PATCH)
patch_opt = lambda url, **kwargs: aopt(patch, url, **kwargs)

options = partial(request, HttpMethods.OPTIONS)
options_opt = lambda url, **kwargs: aopt(options, url, **kwargs)
